const jwt = require("jsonwebtoken");
const bcrypt = require('bcryptjs');
const crypto = require('crypto');

const Config = require("../commons/config");
const User = require('../models/user.model');

const UserErrors = require('../commons/userErrors');
const AuthErrors = require('../commons/authErrors');

const {answer} = require('./ControllerAnswer')

/**
 * check user credentials
 * @param {Object} req - The request object (provided by express)
 * @param {Object} req.body - The data payload sent with the request
 * @param {string} req.body.login - The login of the user to sign in
 * @param {string} req.body.password - The password of the user to sign in
 * @param {Object} res - The result object used to send the result to the client (provided by express)
 * @param {Function} next - The next middleware to call after this one
 * @alias module:AuthController.signIn
 */
const signIn = async function (req, res, next) {
  answer.reset()

  console.log('sign in');

  // sanity check on parameters
  if (!req.body.login) {
    answer.set(AuthErrors.getError(AuthErrors.ERR_AUTH_LOGIN_NOT_DEFINED))
    return next(answer);
  }
  if (!req.body.password) {
    answer.set(AuthErrors.getError(AuthErrors.ERR_AUTH_PASSWORD_NOT_DEFINED))
    return next(answer);
  }
  // check if name exists
  let user = null;
  try {
    user = await User.findOne({login:req.body.login}).exec();
    if (user === null) {
      answer.set(UserErrors.getError(UserErrors.ERR_USER_CANNOT_FIND_LOGIN))
      return next(answer);
    }
  }
  catch(err) {
    answer.set(UserErrors.getError(UserErrors.ERR_USER_INVALID_FIND_LOGIN_REQUEST))
    return next(answer);
  }
  let passValid = bcrypt.compareSync(req.body.password, user.password);
  if (!passValid) {
    answer.set(AuthErrors.getError(AuthErrors.ERR_AUTH_PASSWORD_NO_MATCH))
    return next(answer);
  }
  const xsrfToken = crypto.randomBytes(64).toString('hex');
  console.log("xsrf token = "+xsrfToken);

  let token = jwt.sign({ id: user._id, xsrfToken }, Config.secretKey, {
    expiresIn: Config.jwtExpiration
  });

  console.log("jwt token = "+token);

  res.cookie('access_token', token, {
    maxAge: Config.jwtExpiration*1000,
    httpOnly: true,
    secure: true,
    signed: true,
    sameSite: 'none'
  });

  answer.reset()
  answer.setPayload({
    name: user.login,
    xsrfToken: xsrfToken,
  })
  console.log(answer);

  res.status(200).send(answer);
};

/**
 * check if a jwt token is provided in request header and check it against
 * the secret key
 * @param {Object} req - The request object (provided by express)
 * @param {Object} res - The result object used to send the result to the client (provided by express)
 * @param {Function} next - The next middleware to call after this one
 * @alias module:AuthController.verifyToken
 */
const verifyToken = async function (req, res, next) {

  answer.reset();
  let accessToken = req.signedCookies.access_token;
  console.log(JSON.stringify(req.signedCookies));
  let xsrfToken = req.headers["x-xsrf-token"];

  console.log("verify tokens: jwt="+accessToken+" | xsrf="+xsrfToken);
  if ((!accessToken) || (!xsrfToken)) {
    answer.set(AuthErrors.getError(AuthErrors.ERR_AUTH_NO_TOKEN))
    return next(answer);
  }

  let decoded = '';
  try {
    decoded = jwt.verify(accessToken, Config.secretKey)
  }
  catch(err) {
      if (err instanceof jwt.TokenExpiredError) {
        answer.set(AuthErrors.getError(AuthErrors.ERR_AUTH_TOKEN_EXPIRED))
        return next(answer);
      } else {
        answer.set(AuthErrors.getError(AuthErrors.ERR_AUTH_NOT_AUTHORIZED))
        return next(answer);
      }
  }
  // check if xsrf in header and jwt match
  if (xsrfToken !== decoded.xsrfToken) {
    console.log("decoded and header xsrf tokens differ !")
    answer.set(AuthErrors.getError(AuthErrors.ERR_AUTH_NOT_AUTHORIZED))
    return next(answer);
  }
  // now find if user in in DB and bind it the req
  try {
    let user = await User.findById(decoded.id).exec()
    if (user === null) {
      answer.set(AuthErrors.getError(AuthErrors.ERR_AUTH_NOT_AUTHORIZED))
      return next(answer);
    }
    req.user = user;
  }
  catch(err) {
    answer.set(AuthErrors.getError(AuthErrors.ERR_AUTH_NOT_AUTHORIZED))
    return next(answer);
  }
  return next();
};

exports.verifyToken = verifyToken;
exports.signIn = signIn;
